from typing import Any
from uuid import UUID

from gen_epix.casedb.domain import command, enum, exc, model
from gen_epix.casedb.domain.policy.abac import BaseCaseAbacPolicy
from gen_epix.casedb.domain.service import BaseCaseService as DomainBaseCaseService
from gen_epix.casedb.services.case.base import BaseCaseService
from gen_epix.commondb.domain.enum import RoleSet as CommonRoleSet
from gen_epix.commondb.util import map_paired_elements
from gen_epix.fastapp import BaseUnitOfWork, CrudOperation, CrudOperationSet
from gen_epix.filter import CompositeFilter, Filter, LogicalOperator


def crud(
    self: BaseCaseService, cmd: command.CrudCommand
) -> list[model.Model] | model.Model | list[UUID] | UUID | list[bool] | bool | None:
    """
    Override the base crud method to apply ABAC restrictions and cascade delete
    where necessary
    """
    # Handle no ABAC restrictions
    if any(isinstance(cmd, x) for x in DomainBaseCaseService.NO_ABAC_COMMAND_CLASSES):
        # No ABAC restrictions
        return super(DomainBaseCaseService, self).crud(cmd)  # type: ignore[return-value]

    # Start unit of work and execute all within this scope
    with self.repository.uow() as uow:
        # Metadata commands
        if any(
            isinstance(cmd, x)
            for x in DomainBaseCaseService.ABAC_METADATA_COMMAND_CLASSES
        ):
            return _crud_metadata(self, uow, cmd)  # type: ignore[no-any-return]
        # Data commands
        elif any(
            isinstance(cmd, x) for x in DomainBaseCaseService.ABAC_DATA_COMMAND_CLASSES
        ):
            return _crud_data(self, uow, cmd)
        else:
            raise AssertionError(
                f"Unexpected command {cmd.__class__.__name__} with operation {cmd.operation.value}"
            )


def _crud_metadata(
    self: BaseCaseService,
    uow: BaseUnitOfWork,
    cmd: command.CrudCommand,
) -> Any:
    """Logic for handling metadata commands"""
    # Metadata admin or above: no @ABAC applied
    assert cmd.user
    if cmd.user.roles.intersection(self.role_set_map[CommonRoleSet.GE_REFDATA_ADMIN]):
        # Metadata admin and above have access to all metadata: no ABAC
        # applied, only RBAC
        return _crud_metadata_by_admin(self, uow, cmd)
    return _crud_metadata_by_non_admin(self, uow, cmd)


def _crud_metadata_by_admin(
    self: BaseCaseService, uow: BaseUnitOfWork, cmd: command.CrudCommand
) -> list[model.Model] | model.Model | list[UUID] | UUID | list[bool] | bool | None:
    """Metadata admin command handling, no ABAC applied"""
    _crud_cascade_delete(self, uow, cmd)
    return super(DomainBaseCaseService, self).crud(cmd)  # type:ignore[return-value]


def _crud_metadata_by_non_admin(
    self: BaseCaseService,
    uow: BaseUnitOfWork,
    cmd: command.CrudCommand,
) -> list[model.Model] | model.Model | list[UUID] | UUID | list[bool] | bool | None:
    """Metadata user command handling, ABAC applied"""
    # @ABAC: get case abac
    case_abac = BaseCaseAbacPolicy.get_case_abac_from_command(cmd)

    # Special case: no policy, allows for internal commands to retrieve all
    if not case_abac:
        # No policy: allows for internal commands to retrieve all
        return super(DomainBaseCaseService, self).crud(cmd)  # type:ignore[return-value]

    # Initialise some
    is_read = cmd.operation in CrudOperationSet.READ_OR_EXISTS.value
    is_delete = cmd.operation in CrudOperationSet.DELETE.value
    access_filter: Filter | None = None

    if not is_read:
        # Only read operations are allowed for metadata commands for these
        # users
        raise AssertionError("Unexpected operation")

    if isinstance(cmd, command.CaseTypeCrudCommand):
        valid_case_type_ids = case_abac.get_case_types_with_any_rights()
        access_filter = self._compose_id_filter(("id", valid_case_type_ids))
        # No cascade delete to force conscious decision to delete from other models
        return _crud_with_access_filter(self, uow, cmd, access_filter)

    elif isinstance(cmd, command.CaseTypeSetMemberCrudCommand):
        valid_case_type_ids = case_abac.get_case_types_with_any_rights()
        access_filter = self._compose_id_filter(("case_type_id", valid_case_type_ids))
        return _crud_with_access_filter(self, uow, cmd, access_filter)

    elif isinstance(cmd, command.CaseTypeSetCrudCommand):
        valid_case_type_ids = case_abac.get_case_types_with_any_rights()
        valid_case_type_set_ids: set[UUID] = (
            self._read_association_with_valid_ids(  # type:ignore[assignment]
                command.CaseTypeSetMemberCrudCommand,
                "case_type_set_id",
                "case_type_id",
                valid_ids2=valid_case_type_ids,
                match_all2=is_delete,  # delete requires all case types
                return_type="ids1",
                uow=uow,
                user=cmd.user,
            )
        )
        access_filter = self._compose_id_filter(("id", valid_case_type_set_ids))
        # No cascade delete to force conscious decision to delete from other models
        return _crud_with_access_filter(self, uow, cmd, access_filter)

    elif isinstance(cmd, command.CaseTypeColCrudCommand):
        valid_case_type_col_ids = case_abac.get_case_type_cols_with_any_rights()
        access_filter = self._compose_id_filter(("id", valid_case_type_col_ids))
        # No cascade delete to force conscious decision to delete from other models
        return _crud_with_access_filter(self, uow, cmd, access_filter)

    elif isinstance(cmd, command.CaseTypeColSetMemberCrudCommand):
        valid_case_type_col_ids = case_abac.get_case_type_cols_with_any_rights()
        access_filter = self._compose_id_filter(
            ("case_type_col_id", valid_case_type_col_ids)
        )
        return _crud_with_access_filter(self, uow, cmd, access_filter)

    elif isinstance(cmd, command.CaseTypeColSetCrudCommand):
        # Determine valid case type cols as those with any rights
        valid_case_type_col_ids = case_abac.get_case_type_cols_with_any_rights()
        valid_case_type_col_set_ids: set[UUID] = (
            self._read_association_with_valid_ids(  # type:ignore[assignment]
                command.CaseTypeColSetMemberCrudCommand,
                "case_type_col_set_id",
                "case_type_col_id",
                valid_ids2=valid_case_type_col_ids,
                match_all2=is_delete,  # delete requires all case type cols
                return_type="ids1",
                uow=uow,
                user=cmd.user,
            )
        )
        access_filter = self._compose_id_filter(("id", valid_case_type_col_set_ids))
        # No cascade delete to force conscious decision to delete from other models
        return _crud_with_access_filter(self, uow, cmd, access_filter)

    raise AssertionError("Unexpected operation")


def _crud_data(
    self: BaseCaseService,
    uow: BaseUnitOfWork,
    cmd: command.CrudCommand,
) -> list[model.Model] | model.Model | list[UUID] | UUID | list[bool] | bool | None:
    """Logic for handling data commands"""
    assert cmd.user
    # App admin or above: no @ABAC applied
    if cmd.user.roles.intersection(self.role_set_map[CommonRoleSet.GE_APP_ADMIN]):
        return _crud_data_by_admin(self, uow, cmd)
    return _crud_data_by_non_admin(self, uow, cmd)


def _crud_data_by_admin(
    self: BaseCaseService,
    uow: BaseUnitOfWork,
    cmd: command.CrudCommand,
) -> list[model.Model] | model.Model | list[UUID] | UUID | list[bool] | bool | None:
    """Data admin command handling, no ABAC applied"""
    # Non-ABAC restrictions not enforced anywhere else
    is_create = cmd.operation in CrudOperationSet.CREATE.value
    is_update = cmd.operation in CrudOperationSet.UPDATE.value
    if (is_create or is_update) and isinstance(cmd, command.CaseSetMemberCrudCommand):
        # Verify that the case set and case have the same case type
        self._verify_case_set_member_case_type(
            cmd.user, cmd.get_objs()  # type:ignore[arg-type]
        )

    # Any other operation
    _crud_cascade_delete(self, uow, cmd)
    return super(DomainBaseCaseService, self).crud(cmd)  # type:ignore[return-value]


def _crud_data_by_non_admin(
    self: BaseCaseService,
    uow: BaseUnitOfWork,
    cmd: command.CrudCommand,
) -> list[model.Model] | model.Model | list[UUID] | UUID | list[bool] | bool | None:
    """Data user command handling, ABAC applied"""
    # @ABAC: get case abac
    case_abac = BaseCaseAbacPolicy.get_case_abac_from_command(cmd)

    # Special case: no policy, allows for internal commands to retrieve all
    if case_abac is None:
        # No policy: allows for internal commands to retrieve all
        return super(DomainBaseCaseService, self).crud(cmd)  # type:ignore[return-value]

    # Initialise some
    is_create = cmd.operation in CrudOperationSet.CREATE.value
    is_read = cmd.operation in CrudOperationSet.READ_OR_EXISTS.value
    is_read_all = cmd.operation == CrudOperation.READ_ALL
    is_update = cmd.operation in CrudOperationSet.UPDATE.value
    is_delete = cmd.operation in CrudOperationSet.DELETE.value
    is_delete_all = cmd.operation == CrudOperation.DELETE_ALL
    access_filter: Filter | None = None
    case_sets: list[model.CaseSet]
    cases: list[model.Case]
    case_data_collection_map: dict[UUID, set[UUID]] = {}
    case_set_data_collection_map: dict[UUID, set[UUID]] = {}
    has_access: dict[UUID, set[UUID]] = {}
    assert cmd.user is not None and cmd.user.id is not None

    # Handle each type of command
    if isinstance(cmd, command.CaseSetCrudCommand):
        # Determine valid case types and data collections
        case_set_ids: list[UUID] = cmd.get_obj_ids()  # type: ignore[assignment]
        if is_create:
            # Implemented through separate create case set command
            raise AssertionError("Unexpected operation")
        elif is_read:
            # At least one data collection with read access is required
            retval = self._retrieve_case_sets_with_content_right(
                uow,
                cmd.user.id,
                case_abac,
                enum.CaseRight.READ_CASE_SET,
                case_set_ids=case_set_ids,
                filter=cmd.query_filter,
            )
            return (
                retval[0] if cmd.operation == CrudOperation.READ_ONE else retval  # type: ignore[return-value]
            )
        elif is_update:
            # At least one data collection with write access is required
            case_sets = self._retrieve_case_sets_with_content_right(
                uow,
                cmd.user.id,
                case_abac,
                enum.CaseRight.WRITE_CASE_SET,
                case_set_ids=case_set_ids,
            )
            return super(DomainBaseCaseService, self).crud(cmd)  # type: ignore[return-value]
        elif is_delete:
            # All linked data collections have remove right
            if is_delete_all:
                # Delete all not allowed due to potential large number of case sets
                raise exc.UnauthorizedAuthError(
                    f"Operation {cmd.operation.value} not allowed for case sets for this user"
                )
            # Get all case sets and data collection links
            assert case_set_ids is not None
            case_sets = self.repository.crud(  # type:ignore[assignment]
                uow,
                cmd.user.id,
                model.CaseSet,
                None,
                case_set_ids,
                CrudOperation.READ_SOME,
            )
            case_set_data_collection_map = self._retrieve_case_set_data_collections_map(
                uow,
                cmd.user.id,
                case_set_ids=case_set_ids,
            )
            # Check if the user has access to all data collections of all requested
            # case sets
            for case_set in case_sets:
                data_collection_ids = case_set_data_collection_map.get(
                    case_set.id, set()  # type:ignore[arg-type]
                )
                is_allowed = case_abac.is_allowed(
                    case_set.case_type_id,
                    enum.CaseRight.REMOVE_CASE_SET,
                    True,
                    created_in_data_collection_id=case_set.created_in_data_collection_id,
                    current_data_collection_ids=data_collection_ids,
                )
                if not is_allowed:
                    raise exc.UnauthorizedAuthError(
                        f"User {cmd.user.id} is not allowed to delete case set {case_set.id}"
                    )
            # Delete with cascade
            _crud_cascade_delete(self, uow, cmd)
            return super(DomainBaseCaseService, self).crud(
                cmd
            )  # type:ignore[return-value]
        else:
            raise AssertionError("Unexpected operation")

    elif isinstance(cmd, command.CaseCrudCommand):
        # Determine valid case types and data collections
        case_ids: list[UUID] = cmd.get_obj_ids()  # type: ignore[assignment]
        if is_create | is_read | is_update:
            # Implemented through separate create case set command
            raise AssertionError("Unexpected operation")
        elif is_delete:
            # All linked data collections have remove right
            if is_delete_all:
                # Delete all not allowed due to potential large number of case
                raise exc.UnauthorizedAuthError(
                    f"Operation {cmd.operation.value} not allowed for cases for this user"
                )
            # Get all cases and data collection links
            assert case_ids is not None
            cases = self.repository.crud(  # type:ignore[assignment]
                uow,
                cmd.user.id,
                model.Case,
                None,
                case_ids,
                CrudOperation.READ_SOME,
            )
            case_data_collection_map = self._retrieve_case_data_collections_map(
                uow,
                cmd.user.id,
                case_ids=case_ids,
            )
            # Check if the user has access to all data collections of all requested
            # cases
            for case in cases:
                data_collection_ids = case_data_collection_map.get(
                    case.id, set()  # type:ignore[arg-type]
                )
                is_allowed = case_abac.is_allowed(
                    case.case_type_id,
                    enum.CaseRight.REMOVE_CASE,
                    True,
                    created_in_data_collection_id=case.created_in_data_collection_id,
                    current_data_collection_ids=data_collection_ids,
                )
                if not is_allowed:
                    raise exc.UnauthorizedAuthError(
                        f"User {cmd.user.id} is not allowed to delete case {case.id}"
                    )
            # Delete with cascade
            _crud_cascade_delete(self, uow, cmd)
            return super(DomainBaseCaseService, self).crud(
                cmd
            )  # type:ignore[return-value]
        else:
            raise AssertionError("Unexpected operation")

    elif isinstance(cmd, command.CaseSetMemberCrudCommand):
        # Delete all not allowed due to potential large number of case set members
        if is_delete_all or is_update:
            raise exc.UnauthorizedAuthError(
                f"Operation {cmd.operation.value} not allowed for case set members for this user"
            )

        # Get case set members
        case_set_members: list[model.CaseSetMember]
        if is_create:
            # Must be able to write the case set and read the case
            case_set_members = cmd.get_objs()  # type:ignore[assignment]
        elif is_read_all:
            # Must be able to read or write the case set and read the case
            case_set_members = self.repository.crud(  # type:ignore[assignment]
                uow,
                cmd.user.id,
                model.CaseSetMember,
                None,
                None,
                CrudOperation.READ_ALL,
                filter=cmd.query_filter,
            )
        elif is_read or is_delete:
            # Must be able to read or write the case set and read the case
            case_set_members = self.repository.crud(  # type:ignore[assignment]
                uow,
                cmd.user.id,
                model.CaseSetMember,
                None,
                cmd.get_obj_ids(),
                CrudOperation.READ_SOME,
            )
        elif is_update:
            # Should not be allowed
            raise AssertionError("Update not allowed for case set members")
        else:
            raise AssertionError("Unexpected operation")

        # All operations require read access to the case: retrieve the cases while
        # checking for this read right to determine this
        cases = self._retrieve_cases_with_content_right(
            uow,
            cmd.user.id,
            case_abac,
            enum.CaseRight.READ_CASE,
            case_ids=list({x.case_id for x in case_set_members}),
            filter_content=False,
            on_invalid_case_id=("ignore" if is_read_all or is_delete_all else "raise"),
        )

        # Retrieve the case sets while checking for the correct right(s)
        uq_case_set_ids = {x.case_set_id for x in case_set_members}
        case_sets = self._retrieve_case_sets_with_content_right(
            uow,
            cmd.user.id,
            case_abac,
            enum.CaseRight.READ_CASE_SET,
            case_set_ids=list(uq_case_set_ids),
            on_invalid_case_set_id="ignore",
        )
        if is_delete and not uq_case_set_ids.issubset({x.id for x in case_sets}):
            # Also check the write case set right since not all case sets have the
            # read right
            case_sets += self._retrieve_case_sets_with_content_right(
                uow,
                cmd.user.id,
                case_abac,
                enum.CaseRight.WRITE_CASE_SET,
                case_set_ids=list(uq_case_set_ids),
                on_invalid_case_set_id="ignore",
            )

        # Check if the user has access to all requested case sets
        unauthorized_case_set_ids = uq_case_set_ids - {x.id for x in case_sets}
        if unauthorized_case_set_ids:
            if is_read_all:
                # unauthorized case set ids not applicable, filter out the case set
                # members instead
                case_set_members = [
                    x
                    for x in case_set_members
                    if x.case_set_id not in unauthorized_case_set_ids
                ]
            else:
                unauthorized_case_set_ids_str = ", ".join(
                    [str(x) for x in unauthorized_case_set_ids]
                )
                raise exc.UnauthorizedAuthError(
                    f"User {cmd.user.id} does not have access to case set(s): {unauthorized_case_set_ids_str}"
                )

        # Execute command in case of create or delete, return case set
        # members in case of read
        if is_create or is_delete:
            return super(DomainBaseCaseService, self).crud(
                cmd
            )  # type:ignore[return-value]
        elif is_read:
            return (
                case_set_members[0]  # type:ignore[return-value]
                if cmd.operation == CrudOperation.READ_ONE
                else case_set_members
            )
        else:
            raise AssertionError("Unexpected operation")

    elif isinstance(cmd, command.CaseDataCollectionLinkCrudCommand):
        # Read all without filter and delete all not allowed due to potential large
        # number of case data collection links
        if (is_read_all and not cmd.query_filter) or is_delete_all or is_update:
            raise exc.UnauthorizedAuthError(
                f"Operation {cmd.operation.value} not allowed for case data collection links for this user"
            )

        # Get case data collection links
        case_data_collection_links: list[model.CaseDataCollectionLink]
        if is_create:
            # Must be able to add the case to all the data collections
            case_data_collection_links = cmd.get_objs()  # type:ignore[assignment]
        elif is_read_all:
            # Must be able to read or write the case in all the data collections
            case_data_collection_links = (
                self.repository.crud(  # type:ignore[assignment]
                    uow,
                    cmd.user.id,
                    model.CaseDataCollectionLink,
                    None,
                    None,
                    CrudOperation.READ_ALL,
                    filter=cmd.query_filter,
                )
            )
        elif is_read or is_delete:
            # Must be able to read or write the case in (for is_read), or remove from
            # (for is_delete), all the data collections
            case_data_collection_links = (
                self.repository.crud(  # type:ignore[assignment]
                    uow,
                    cmd.user.id,
                    model.CaseDataCollectionLink,
                    None,
                    cmd.get_obj_ids(),
                    CrudOperation.READ_SOME,
                )
            )
        elif is_update:
            # Should not be allowed
            raise AssertionError("Update not allowed for case data collection links")
        else:
            raise AssertionError("Unexpected operation")

        # Go over each case and check if the user has the required rights to it
        case_data_collection_map = map_paired_elements(  # type: ignore[assignment]
            ((x.case_id, x.data_collection_id) for x in case_data_collection_links),
            as_set=True,
        )
        uq_case_ids = set(case_data_collection_map.keys())
        cases = self.repository.crud(  # type:ignore[assignment]
            uow,
            cmd.user.id,
            model.Case,
            None,
            list(uq_case_ids),
            CrudOperation.READ_SOME,
        )
        if is_read:
            # Get read or write access per case type
            has_read_access = case_abac.get_combinations_with_access_right(
                enum.CaseRight.READ_CASE
            )
            has_write_access = case_abac.get_combinations_with_access_right(
                enum.CaseRight.WRITE_CASE
            )
            has_access = {x: set(y) for x, y in has_read_access.items()}
            for x, y in has_write_access.items():
                if x in has_access:
                    has_access[x].update(y)
                else:
                    has_access[x] = y
        for case in cases:
            case_type_id = case.case_type_id
            created_in_data_collection_id = case.created_in_data_collection_id
            data_collection_ids = case_data_collection_map[
                case.id  # type:ignore[index]
            ]
            if is_read:
                if not data_collection_ids.intersection(
                    has_access.get(case_type_id, set())
                ):
                    raise exc.UnauthorizedAuthError(
                        f"User {cmd.user.id} does not have read or write access to case {case.id} in all data collections"
                    )
            elif is_create:
                is_allowed = case_abac.is_allowed(
                    case_type_id,
                    enum.CaseRight.ADD_CASE,
                    False,
                    created_in_data_collection_id=created_in_data_collection_id,
                    current_data_collection_ids=data_collection_ids,
                    tgt_data_collection_ids=data_collection_ids,
                )
                if not is_allowed:
                    raise exc.UnauthorizedAuthError(
                        f"User {cmd.user.id} does not have add access to case {case.id} to all data collections"
                    )
            elif is_delete:
                is_allowed = case_abac.is_allowed(
                    case_type_id,
                    enum.CaseRight.REMOVE_CASE,
                    False,
                    created_in_data_collection_id=created_in_data_collection_id,
                    current_data_collection_ids=data_collection_ids,
                    tgt_data_collection_ids=data_collection_ids,
                )
                if not is_allowed:
                    raise exc.UnauthorizedAuthError(
                        f"User {cmd.user.id} does not have remove access to case {case.id} to all data collections"
                    )

        # Execute command in case of create or delete, return case data collection
        # links in case of read
        if is_create or is_delete:
            return super(DomainBaseCaseService, self).crud(
                cmd
            )  # type:ignore[return-value]
        elif is_read:
            return (
                case_data_collection_links[0]  # type:ignore[return-value]
                if cmd.operation == CrudOperation.READ_ONE
                else case_data_collection_links
            )
        else:
            raise AssertionError("Unexpected operation")

    elif isinstance(cmd, command.CaseSetDataCollectionLinkCrudCommand):
        # Read all without filter and delete all not allowed due to potential large
        # number of case set data collection links
        if (is_read_all and not cmd.query_filter) or is_delete_all or is_update:
            raise exc.UnauthorizedAuthError(
                f"Operation {cmd.operation.value} not allowed for case set data collection links for this user"
            )

        # Get case set data collection links
        case_set_data_collection_links: list[model.CaseSetDataCollectionLink]
        if is_create:
            # Must be able to add the case set to all the data collections
            case_set_data_collection_links = cmd.get_objs()  # type:ignore[assignment]
        elif is_read_all:
            # Must be able to read or write the case set in all the data collections
            case_set_data_collection_links = (
                self.repository.crud(  # type:ignore[assignment]
                    uow,
                    cmd.user.id,
                    model.CaseSetDataCollectionLink,
                    None,
                    None,
                    CrudOperation.READ_ALL,
                    filter=cmd.query_filter,
                )
            )
        elif is_read or is_delete:
            # Must be able to read or write the case set in (for is_read), or remove from
            # (for is_delete), all the data collections
            case_set_data_collection_links = (
                self.repository.crud(  # type:ignore[assignment]
                    uow,
                    cmd.user.id,
                    model.CaseSetDataCollectionLink,
                    None,
                    cmd.get_obj_ids(),
                    CrudOperation.READ_SOME,
                )
            )
        elif is_update:
            # Should not be allowed
            raise AssertionError(
                "Update not allowed for case set data collection links"
            )
        else:
            raise AssertionError("Unexpected operation")

        # Go over each case set and check if the user has the required rights to it
        case_set_data_collection_map = map_paired_elements(  # type:ignore[assignment]
            (
                (x.case_set_id, x.data_collection_id)
                for x in case_set_data_collection_links
            ),
            as_set=True,
        )
        uq_case_set_ids = set(case_set_data_collection_map.keys())
        case_sets = self.repository.crud(  # type:ignore[assignment]
            uow,
            cmd.user.id,
            model.CaseSet,
            None,
            list(uq_case_set_ids),
            CrudOperation.READ_SOME,
        )
        if is_read:
            # Get read or write access per case type
            has_read_access = case_abac.get_combinations_with_access_right(
                enum.CaseRight.READ_CASE_SET
            )
            has_write_access = case_abac.get_combinations_with_access_right(
                enum.CaseRight.WRITE_CASE_SET
            )
            has_access = {x: set(y) for x, y in has_read_access.items()}
            for x, y in has_write_access.items():
                if x in has_access:
                    has_access[x].update(y)
                else:
                    has_access[x] = y
        for case_set in case_sets:
            case_type_id = case_set.case_type_id
            created_in_data_collection_id = case_set.created_in_data_collection_id
            data_collection_ids = case_set_data_collection_map[
                case_set.id  # type:ignore[index]
            ]
            if is_read:
                if not data_collection_ids.intersection(
                    has_access.get(case_type_id, set())
                ):
                    raise exc.UnauthorizedAuthError(
                        f"User {cmd.user.id} does not have read or write access to case set {case_set.id} in all data collections"
                    )
            elif is_create:
                is_allowed = case_abac.is_allowed(
                    case_type_id,
                    enum.CaseRight.ADD_CASE_SET,
                    False,
                    created_in_data_collection_id=created_in_data_collection_id,
                    current_data_collection_ids=data_collection_ids,
                    tgt_data_collection_ids=data_collection_ids,
                )
                if not is_allowed:
                    raise exc.UnauthorizedAuthError(
                        f"User {cmd.user.id} does not have add access to case set {case_set.id} to all data collections"
                    )
            elif is_delete:
                is_allowed = case_abac.is_allowed(
                    case_type_id,
                    enum.CaseRight.REMOVE_CASE_SET,
                    False,
                    created_in_data_collection_id=created_in_data_collection_id,
                    current_data_collection_ids=data_collection_ids,
                    tgt_data_collection_ids=data_collection_ids,
                )
                if not is_allowed:
                    raise exc.UnauthorizedAuthError(
                        f"User {cmd.user.id} does not have remove access to case set {case_set.id} to all data collections"
                    )

        # Execute command in case of create or delete, return case set data
        # collection links in case of read
        if is_create or is_delete:
            return super(DomainBaseCaseService, self).crud(
                cmd
            )  # type:ignore[return-value]
        elif is_read:
            return (
                case_set_data_collection_links[0]  # type:ignore[return-value]
                if cmd.operation == CrudOperation.READ_ONE
                else case_set_data_collection_links
            )
        else:
            raise AssertionError("Unexpected operation")

    raise AssertionError("Unexpected operation")


def _crud_cascade_delete(
    self: BaseCaseService, uow: BaseUnitOfWork, cmd: command.CrudCommand
) -> None:
    """
    In case of a delete operation, cascade delete all instances of any
    linked_model_classes that are linked to the instances in cmd.
    """
    is_delete = cmd.operation in CrudOperationSet.DELETE.value
    if not is_delete:
        # Not a delete opertion: nothing to do
        return
    model_class = cmd.MODEL_CLASS
    link_model_classes: list[type[model.Model]] | None = None
    for (
        model_base_class,
        link_model_classes_tuple,
    ) in BaseCaseService.CASCADE_DELETE_MODEL_CLASSES.items():
        if issubclass(model_class, model_base_class):
            link_model_classes = list(link_model_classes_tuple)
            break
    if link_model_classes is None:
        # No cascade delete: nothing to do
        return
    assert cmd.user is not None and cmd.user.id is not None
    obj_ids: set[UUID] | None = cmd.get_obj_ids(as_set=True)  # type:ignore[assignment]
    # Go over each link_model_class and delete all instances that are linked to
    # the instances in cmd
    for link_model_class in link_model_classes:
        assert link_model_class.ENTITY is not None
        get_obj_id = link_model_class.ENTITY.get_obj_id
        if cmd.operation == CrudOperation.DELETE_ALL:
            # Special case: delete all instances
            self.repository.crud(
                uow,
                cmd.user.id,
                link_model_class,
                None,
                None,
                CrudOperation.DELETE_ALL,
            )
            continue
        assert obj_ids is not None
        # Get the instances that are linked to the instances in cmd
        get_link_id = link_model_class.ENTITY.get_link_id(model_class)
        link_objs: list = self.repository.crud(  # type:ignore[assignment]
            uow,
            cmd.user.id,
            link_model_class,
            None,
            None,
            CrudOperation.READ_ALL,
        )
        link_obj_ids = [get_obj_id(x) for x in link_objs if get_link_id(x) in obj_ids]
        # Delete these instances
        self.repository.crud(
            uow,
            cmd.user.id,
            link_model_class,
            None,
            link_obj_ids,
            CrudOperation.DELETE_SOME,
        )


def _crud_with_access_filter(
    self: BaseCaseService,
    uow: BaseUnitOfWork,
    cmd: command.CrudCommand,
    access_filter: Filter | None = None,
    cascade_if_delete: bool = False,
) -> list[model.Model] | model.Model | list[UUID] | UUID | list[bool] | bool | None:
    # Set access filter if any and call generic crud
    orig_access_filter = cmd.access_filter
    if access_filter:
        if cmd.access_filter:
            cmd.access_filter = CompositeFilter(
                filters=[access_filter, cmd.access_filter],  # type: ignore[list-item]
                operator=LogicalOperator.AND,
            )
        else:
            cmd.access_filter = access_filter
    if cascade_if_delete:
        _crud_cascade_delete(self, uow, cmd)
    retval = super(DomainBaseCaseService, self).crud(cmd)
    cmd.access_filter = orig_access_filter
    return retval  # type:ignore[return-value]



from uuid import UUID

import gen_epix.casedb.domain.command as command
import gen_epix.casedb.domain.enum as enum
import gen_epix.casedb.domain.model as model
import gen_epix.seqdb.domain.command as seqdb_command
from gen_epix.casedb.domain import exc
from gen_epix.casedb.domain.policy import BaseCaseAbacPolicy
from gen_epix.casedb.domain.service import BaseCaseService as DomainBaseCaseService
from gen_epix.casedb.services.case.base import BaseCaseService
from gen_epix.fastapp import CrudOperation
from gen_epix.fastapp.unit_of_work import BaseUnitOfWork


def case_service_create_read_sets_or_seqs_for_cases(
    self: BaseCaseService,
    cmd: command.CreateReadSetsForCasesCommand | command.CreateSeqsForCasesCommand,
) -> list[model.ReadSet] | list[model.Seq]:
    user, repository = self._get_user_and_repository(cmd)
    assert isinstance(user, model.User) and user.id is not None

    # Parse input
    read_sets: list[model.ReadSet] = []
    seqs: list[model.Seq] = []
    case_ids: list[UUID] = []
    case_type_col_ids: list[UUID] = []
    if isinstance(cmd, command.CreateReadSetsForCasesCommand):
        is_read_set = True
        read_sets = [x.read_set for x in cmd.case_read_sets]  # type:ignore
        case_ids = [x.case_id for x in cmd.case_read_sets]
        case_type_col_ids = [x.case_type_col_id for x in cmd.case_read_sets]
    elif isinstance(cmd, command.CreateSeqsForCasesCommand):
        is_read_set = False
        seqs = [x.seq for x in cmd.case_seqs]  # type:ignore
        case_ids = [x.case_id for x in cmd.case_seqs]
        case_type_col_ids = [x.case_type_col_id for x in cmd.case_seqs]
    else:
        raise exc.InvalidArgumentsError("Invalid command type")

    # Special case: nothing to create
    if is_read_set and not read_sets:
        return []
    if not is_read_set and not seqs:
        return []

    # Retrieve case ABAC
    case_abac = BaseCaseAbacPolicy.get_case_abac_from_command(cmd)
    assert case_abac is not None

    # Handle transactions
    with repository.uow() as uow:
        cases = _get_cases_for_create_read_sets_or_seqs(
            self, cmd, case_abac, uow, user.id, case_ids, case_type_col_ids
        )

        # Create ReadSets or Seqs
        created_objs: list[model.ReadSet] | list[model.Seq]
        command_class = (
            seqdb_command.ReadSetCrudCommand
            if is_read_set
            else seqdb_command.SeqCrudCommand
        )
        created_objs = self.app.handle(
            command_class(
                user=cmd.user,
                operation=CrudOperation.CREATE_SOME,
                objs=read_sets if is_read_set else seqs,  # type: ignore[arg-type]
            )
        )

        # Update Cases with created ReadSet or Seq IDs
        for case, case_type_col_id, created_obj in zip(
            cases, case_type_col_ids, created_objs
        ):
            case.content[case_type_col_id] = str(created_obj.id)
        super(DomainBaseCaseService, self).crud(
            command.CaseCrudCommand(
                user=cmd.user,
                operation=CrudOperation.UPDATE_SOME,
                objs=cases,  # type: ignore[arg-type]
            )
        )

    return created_objs


def case_service_create_file_for_read_set_or_seq(
    self: BaseCaseService,
    cmd: command.CreateFileForReadSetCommand | command.CreateFileForSeqCommand,
) -> UUID:
    user, repository = self._get_user_and_repository(cmd)
    user_id: UUID = user.id  # type: ignore[assignment]

    if isinstance(cmd, command.CreateFileForReadSetCommand):
        is_read_set = True
    elif isinstance(cmd, command.CreateFileForSeqCommand):
        is_read_set = False
    else:
        raise exc.InvalidArgumentsError("Invalid command type")

    # Retrieve case ABAC
    case_abac = BaseCaseAbacPolicy.get_case_abac_from_command(cmd)
    assert case_abac is not None

    # Handle transaction for reading case
    with repository.uow() as uow:
        cases = _get_cases_for_create_read_sets_or_seqs(
            self, cmd, case_abac, uow, user_id, [cmd.case_id], [cmd.case_type_col_id]
        )
        case = cases[0]

        # Retrieve ReadSet or Seq ID from case content
        if cmd.case_type_col_id not in case.content:
            raise exc.InvalidArgumentsError(
                "No ReadSet linked to case for the given case type column"
            )
        read_set_or_seq_id = UUID(case.content[cmd.case_type_col_id])

    # Retrieve ReadSet or Seq, verify no file linked yet, and create file
    def _create_file(
        self: BaseCaseService,
        cmd: command.CreateFileForReadSetCommand | command.CreateFileForSeqCommand,
    ) -> model.File:
        created_file: model.File = self.app.handle(
            seqdb_command.FileCrudCommand(
                user=cmd.user,
                operation=CrudOperation.CREATE_ONE,
                objs=model.File(content=cmd.file_content),
            )
        )
        return created_file

    if is_read_set:
        assert isinstance(cmd, command.CreateFileForReadSetCommand)
        # Verify no file linked yet
        read_set: model.ReadSet = self.app.handle(
            seqdb_command.ReadSetCrudCommand(
                user=cmd.user,
                operation=CrudOperation.READ_ONE,
                obj_ids=read_set_or_seq_id,
            )
        )
        if cmd.is_fwd and read_set.fwd_file_id is not None:
            raise exc.InvalidArgumentsError(
                "The ReadSet already has a forward file linked"
            )
        if not cmd.is_fwd and read_set.rev_file_id is not None:
            raise exc.InvalidArgumentsError(
                "The ReadSet already has a reverse file linked"
            )
        # Create Seq
        created_file = _create_file(self, cmd)
        # Update ReadSet with file ID
        if cmd.is_fwd:
            read_set.fwd_file_id = created_file.id
        else:
            read_set.rev_file_id = created_file.id
        self.app.handle(
            seqdb_command.ReadSetCrudCommand(
                user=cmd.user,
                operation=CrudOperation.UPDATE_ONE,
                objs=read_set,
            )
        )
    else:
        assert isinstance(cmd, command.CreateFileForSeqCommand)
        # Verify no file linked yet
        seq: model.Seq = self.app.handle(
            seqdb_command.SeqCrudCommand(
                user=cmd.user,
                operation=CrudOperation.READ_ONE,
                obj_ids=read_set_or_seq_id,
            )
        )
        if seq.file_id is not None:
            raise exc.InvalidArgumentsError("The Seq already has a file linked")
        # Create file
        created_file = _create_file(self, cmd)
        # Update Seq with file ID
        seq.file_id = created_file.id
        self.app.handle(
            seqdb_command.SeqCrudCommand(
                user=cmd.user,
                operation=CrudOperation.UPDATE_ONE,
                objs=seq,
            )
        )

    assert created_file.id is not None
    return created_file.id


def _get_cases_for_create_read_sets_or_seqs(
    self: BaseCaseService,
    cmd: (
        command.CreateReadSetsForCasesCommand
        | command.CreateSeqsForCasesCommand
        | command.CreateFileForReadSetCommand
        | command.CreateFileForSeqCommand
    ),
    case_abac: model.CaseAbac,
    uow: BaseUnitOfWork,
    user_id: UUID,
    case_ids: list[UUID],
    case_type_col_ids: list[UUID],
) -> list[model.Case]:
    # Get CaseTypeCol and col data
    case_type_cols: list[model.CaseTypeCol] = (
        self.repository.crud(  # type:ignore[assignment]
            uow,
            user_id,
            model.CaseTypeCol,
            None,
            list(set(case_type_col_ids)),
            CrudOperation.READ_SOME,
        )
    )
    case_type_col_by_id: dict[UUID, model.CaseTypeCol] = {
        x.id: x for x in case_type_cols if x.id is not None
    }

    # Get Col data
    col_ids: set[UUID] = {case_type_col.col_id for case_type_col in case_type_cols}
    cols: list[model.Col] = self.repository.crud(  # type: ignore[assignment]
        uow,
        user_id,
        model.Col,
        None,
        list(col_ids),
        CrudOperation.READ_SOME,
    )
    col_by_id: dict[UUID, model.Col] = {x.id: x for x in cols if x.id is not None}

    # TODO: Verify all case type cols are for the given case type

    # Verify all case type cols are of type GENETIC_READS
    if isinstance(
        cmd,
        (command.CreateReadSetsForCasesCommand, command.CreateFileForReadSetCommand),
    ):
        expected_col_type = enum.ColType.GENETIC_READS
    elif isinstance(
        cmd, (command.CreateSeqsForCasesCommand, command.CreateFileForSeqCommand)
    ):
        expected_col_type = enum.ColType.GENETIC_SEQUENCE
    else:
        raise exc.InvalidArgumentsError("Invalid command type")
    invalid_case_type_col_ids = [
        x.col_id
        for x in case_type_cols
        if col_by_id[x.col_id].col_type != expected_col_type
    ]
    if invalid_case_type_col_ids:
        invalid_case_type_col_ids_str = ", ".join(
            str(x) for x in invalid_case_type_col_ids
        )
        raise exc.InvalidArgumentsError(
            f"Some columns are not of type GENETIC_READS: {invalid_case_type_col_ids_str}"
        )

    # Get Case data
    cases: list[model.Case] = self.repository.crud(  # type:ignore[assignment]
        uow,
        user_id,
        model.Case,
        None,
        list(case_ids),
        CrudOperation.READ_SOME,
    )

    # Verify if all case type cols are for the same case type as the cases
    # TODO: remove once the command is adjusted to be for only one case type
    invalid_case_type_col_ids = [
        y
        for x, y in zip(cases, case_type_col_ids)
        if case_type_col_by_id[y].case_type_id != x.case_type_id
    ]
    if invalid_case_type_col_ids:
        invalid_case_type_col_ids_str = ", ".join(
            str(x) for x in invalid_case_type_col_ids
        )
        raise exc.InvalidArgumentsError(
            f"Some case type column ids are for a different case type than the given cases: {invalid_case_type_col_ids_str}"
        )

    # @ABAC: Verify write rights to case_type_col for each case
    if not case_abac.is_full_access:
        # Get some write rights to case_type_col
        writable_data_collections_by_case_type_col: dict[UUID, set[UUID]] = {
            x: case_abac.get_data_collections_with_access_right_for_case_type_col(
                x, enum.CaseRight.WRITE_CASE
            )
            for x in case_type_col_ids
        }
        # Retrieve data collections by case id for ABAC column-level write checks
        case_data_collections_map: dict[UUID, set[UUID]] = (
            self._retrieve_case_data_collections_map(uow, user_id, case_ids=case_ids)
        )
        # For each requested (case, case_type_col), ensure the user has write access
        # to that column in at least one data collection the case belongs to
        for case, case_type_col_id in zip(cases, case_type_col_ids):
            # Membership includes created_in_data_collection_id
            assert case.id is not None
            case_data_collections = set(case_data_collections_map.get(case.id, set()))
            case_data_collections.add(case.created_in_data_collection_id)
            if case_data_collections.isdisjoint(
                writable_data_collections_by_case_type_col[case_type_col_id]
            ):
                raise exc.UnauthorizedAuthError(
                    "User has no WRITE_CASE access to the specified column in any data collection of the case"
                )
    return cases



from gen_epix.fastapp.repositories import DictRepository
from gen_epix.omopdb.domain.repository.omop import BaseOmopRepository


class OmopDictRepository(DictRepository, BaseOmopRepository):
    pass



from enum import Enum


class Role(Enum):
    ROOT = "COMMONDB_ROOT"
    APP_ADMIN = "COMMONDB_APP_ADMIN"
    ORG_ADMIN = "COMMONDB_ORG_ADMIN"
    REFDATA_ADMIN = "COMMONDB_REFDATA_ADMIN"
    ORG_USER = "COMMONDB_ORG_USER"
    GUEST = "COMMONDB_GUEST"
    ROLE1 = "COMMONDB_ROLE1"


class RoleSet(Enum):
    ALL = frozenset(
        {
            Role.ROOT,
            Role.APP_ADMIN,
            Role.ORG_ADMIN,
            Role.REFDATA_ADMIN,
            Role.ORG_USER,
            Role.GUEST,
        }
    )
    GE_APP_ADMIN = frozenset({Role.ROOT, Role.APP_ADMIN})
    GE_ORG_ADMIN = frozenset({Role.ROOT, Role.APP_ADMIN, Role.ORG_ADMIN})
    GE_REFDATA_ADMIN = frozenset({Role.ROOT, Role.APP_ADMIN, Role.REFDATA_ADMIN})
    GE_ORG_USER = frozenset(
        {
            Role.ROOT,
            Role.APP_ADMIN,
            Role.ORG_ADMIN,
            Role.ORG_USER,
        }
    )
    GE_GUEST = frozenset(
        {
            Role.ROOT,
            Role.APP_ADMIN,
            Role.ORG_ADMIN,
            Role.ORG_USER,
            Role.GUEST,
        }
    )
    LT_ORG_ADMIN = frozenset({Role.ORG_USER, Role.GUEST})
    ROOT = frozenset({Role.ROOT})
    APPLICATION = frozenset({Role.APP_ADMIN})
    ORGANIZATION = frozenset({Role.APP_ADMIN, Role.ORG_ADMIN})
    METADATA = frozenset({Role.REFDATA_ADMIN})
    OPERATIONAL = frozenset({Role.ORG_USER, Role.GUEST})


class ServiceType(Enum):
    AUTH = "AUTH"
    ORGANIZATION = "ORGANIZATION"
    SYSTEM = "SYSTEM"
    RBAC = "RBAC"
    ABAC = "ABAC"


class RepositoryType(Enum):
    DICT = "DICT"
    SA_SQLITE = "SA_SQLITE"
    SA_SQL = "SA_SQL"


class AppType(Enum):
    COMMONDB = "COMMONDB"
    CASEDB = "CASEDB"
    SEQDB = "SEQDB"
    OMOPDB = "OMOPDB"
    ALL = "ALL"


class AppTypeSet(Enum):
    ALL = frozenset({AppType.COMMONDB, AppType.CASEDB, AppType.SEQDB, AppType.OMOPDB})


class AppConfigType(Enum):
    IDPS = "idps"
    MOCK_IDPS = "mock_idps"
    NO_AUTH = "no_auth"
    DEBUG = "debug"


class DevIdpConfig(Enum):
    IDPS = "IDPS"
    MOCK = "MOCK"
    NONE = "NONE"


class DevRepositoryConfig(Enum):
    DICT_DEMO = "DICT_DEMO"
    DICT_EMPTY = "DICT_EMPTY"
    SA_SQLITE_DEMO = "SA_SQLITE_DEMO"
    SA_SQLITE_EMPTY = "SA_SQLITE_EMPTY"
    SA_SQL = "SA_SQL"


class DevRepositoryConfigSet(Enum):
    DICT = frozenset({DevRepositoryConfig.DICT_DEMO, DevRepositoryConfig.DICT_EMPTY})
    SA = frozenset(
        {
            DevRepositoryConfig.SA_SQLITE_DEMO,
            DevRepositoryConfig.SA_SQLITE_EMPTY,
            DevRepositoryConfig.SA_SQL,
        }
    )
    SA_SQLITE = frozenset(
        {DevRepositoryConfig.SA_SQLITE_DEMO, DevRepositoryConfig.SA_SQLITE_EMPTY}
    )
    SA_SQL = frozenset({DevRepositoryConfig.SA_SQL})
    DEMO = frozenset(
        {DevRepositoryConfig.DICT_DEMO, DevRepositoryConfig.SA_SQLITE_DEMO}
    )
    EMPTY = frozenset(
        {DevRepositoryConfig.DICT_EMPTY, DevRepositoryConfig.SA_SQLITE_EMPTY}
    )



import logging
import uuid
from collections.abc import Callable, Hashable
from functools import partial
from typing import Any, NoReturn

from gen_epix.commondb.domain import model
from gen_epix.fastapp import App, LogLevel, exc
from gen_epix.fastapp.api import exc as api_exc

http_exception_fmap = {
    400: api_exc.BadRequest400HTTPException,
    401: api_exc.UnauthorizedUser401HTTPException,
    403: api_exc.Forbidden403HTTPException,
    404: api_exc.ResourceNotFound404HTTPException,
    405: api_exc.MethodNotAllowed405HTTPException,
    409: api_exc.ResourceConflict409HTTPException,
    422: api_exc.UnprocessableEntity422HTTPException,
    500: api_exc.InternalServerError500HTTPException,
    503: api_exc.ServiceUnavailableError503HTTPException,
}


def get_logger_fmap(logger: logging.Logger) -> dict[LogLevel, Callable]:
    logger_fmap = {
        LogLevel.TRACE: logger.debug,
        LogLevel.DEBUG: logger.debug,
        LogLevel.INFO: logger.info,
        LogLevel.WARN: logger.warning,
        LogLevel.ERROR: logger.error,
        LogLevel.FATAL: logger.critical,
    }
    return logger_fmap


# For debugging purposes
LAST_HANDLED_EXCEPTION: dict[str, Any] = {
    "id": uuid.uuid4(),
}


# TODO: Consider refactoring this into a callable ExceptionHandler class
def generate_handle_exception_function(
    app: App,
    logger: logging.Logger | None,
) -> Callable[
    [str, model.User | None, Exception, Hashable | list[Hashable] | None],
    NoReturn,
]:

    def handle_exception(
        app: App,
        logger: logging.Logger | None,
        log_message_id: str,
        user: model.User | None,
        exception: Exception,
        request_ids: Hashable | list[Hashable] | None = None,
        level: LogLevel = LogLevel.ERROR,
    ) -> NoReturn:
        LAST_HANDLED_EXCEPTION.update(
            {
                "id": uuid.uuid4(),
                "log_message_id": log_message_id,
                "user": user,
                "exception": exception,
                "request_ids": request_ids,
                "level": level,
            }
        )
        # Log without stack trace since this is expected to be logged separately
        log_message = app.create_log_message(
            log_message_id, None, user_id=user.id if user else None, exception=exception
        )
        # Raise HTTP exception
        if isinstance(exception, exc.DomainException):
            if isinstance(exception, exc.IdsError):
                http_status_code = 422
                if isinstance(
                    exception, (exc.LinkConstraintViolationError, exc.DuplicateIdsError)
                ):
                    http_status_code = 409
                invalid_ids = []
                if request_ids and exception.ids:
                    # Compare ids received in request with those reported
                    # as invalid in the DomainError
                    raw_request_ids = request_ids
                    if not isinstance(raw_request_ids, list):
                        raw_request_ids = [raw_request_ids]
                    request_ids = []
                    for id_ in raw_request_ids:
                        if not id_:
                            continue
                        if isinstance(id_, list):
                            request_ids += [x for x in id_ if x]
                        else:
                            request_ids.append(id_)
                    invalid_ids = [x for x in request_ids if x in exception.ids]
                if invalid_ids:
                    # (Part of the) issue is with id(s). Provide detail on that.
                    if isinstance(exception, exc.DuplicateIdsError):
                        invalid_ids_str = ", ".join(
                            [f'"{x}"' for x in set(invalid_ids)]
                        )
                        detail = f"Duplicate ids(s) provided: {invalid_ids_str}"
                    else:
                        invalid_ids_str = ", ".join([f'"{x}"' for x in invalid_ids])
                        detail = f"Invalid ids(s) provided: {invalid_ids_str}"
                    if logger:
                        logger.info(log_message)
                    raise http_exception_fmap[http_status_code](
                        detail=detail
                    ) from exception
                # IdsError, but other issue than provided invalid ids.
                # No further details provided.
                if logger:
                    logger.info(log_message)
                raise http_exception_fmap[http_status_code]() from exception
            elif isinstance(exception, exc.AuthException):
                if logger:
                    logger.info(log_message)
                raise http_exception_fmap[exception.get_http_status_code()](
                    detail="Access denied", **exception.get_http_other_props()
                ) from exception
            elif isinstance(exception, exc.ServiceException):
                if logger:
                    logger.error(log_message)
                raise http_exception_fmap[exception.get_http_status_code()](
                    detail="System unavailable", **exception.get_http_other_props()
                ) from exception
            else:
                # Other domain issue.
                if logger:
                    logger.warning(log_message)
                raise http_exception_fmap[422](detail=str(exception)) from exception
        else:
            # Any other error than a DomainError
            if logger:
                logger.error(log_message)
            raise http_exception_fmap[500]() from exception

    return partial(handle_exception, app, logger)



"""
Core transformer classes and interfaces.
"""

from abc import ABC, abstractmethod
from typing import Any

from gen_epix.transform.adapter import ObjectAdapter
from gen_epix.transform.transform_result import TransformResult


class Transformer(ABC):
    """Base transformer class for single object transformations."""

    def __init__(self, name: str | None = None):
        self.name = name or self.__class__.__name__

    @abstractmethod
    def transform(self, obj: ObjectAdapter) -> ObjectAdapter:
        """
        Transform a single object.

        Args:
            obj: Object adapter wrapping the object to transform

        Returns:
            Transformed object adapter
        """
        pass

    def __call__(self, obj: Any) -> TransformResult:
        """
        Transform an object with error handling.

        Args:
            obj: Object to transform

        Returns:
            TransformResult containing success/failure information
        """
        try:
            adapter = ObjectAdapter(obj)
            transformed_adapter = self.transform(adapter)
            return TransformResult(
                success=True,
                original_object=obj,
                transformed_object=transformed_adapter.unwrap(),
                transformer_name=self.name,
            )
        except Exception as e:
            return TransformResult(
                success=False, original_object=obj, error=e, transformer_name=self.name
            )



from collections.abc import Callable, Hashable, Iterable
from typing import Any, Literal

from gen_epix.filter.base import Filter
from gen_epix.filter.enum import FilterType


class ExistsFilter(Filter):

    def match_value(
        self,
        value: Any | None,
        na_values: set[Any] | None = None,
        map_fn: Callable[[Any], Any] | None = None,
    ) -> bool:
        if na_values is None:
            return (value is not None) ^ self.invert
        return (value not in na_values) ^ self.invert

    def match_column(
        self,
        values: Iterable[Any | None],
        na_values: set[Any] | None = None,
        map_fn: Callable[[Any], Any] | None = None,
    ) -> Iterable[bool]:
        if na_values is None:
            for value in values:
                yield (value is not None) ^ self.invert
        else:
            for value in values:
                yield (value not in na_values) ^ self.invert

    def match_row(
        self,
        row: dict[Hashable, Any | None],
        na_values: set[Any] | None = None,
        map_fn: Callable[[Any], Any] | None = None,
    ) -> bool:
        if self.key is None:
            raise ValueError("Key must be set to apply filter to a row.")
        # Match if both key exists and value not null
        key = self.key
        if na_values is None:
            return ((key in row) and (row[key] is not None)) ^ self.invert
        return ((key in row) and (row[key] not in na_values)) ^ self.invert

    def match_rows(
        self,
        rows: Iterable[dict[Hashable, Any | None]],
        na_values: set[Any] | None = None,
        map_fn: Callable[[Any], Any] | None = None,
    ) -> Iterable[bool]:
        if self.key is None:
            raise ValueError("Key must be set to apply filter to a row.")
        # Match if both key exists and value not null
        key = self.key
        if na_values is None:
            for row in rows:
                yield ((key in row) and (row[key] is not None)) ^ self.invert
        else:
            for row in rows:
                yield ((key in row) and (row[key] not in na_values)) ^ self.invert

    def _match(self, value: Any) -> bool:
        return True


class TypedExistsFilter(ExistsFilter):
    type: Literal[FilterType.EXISTS.value]



from typing import Any, Literal, Self
from uuid import UUID

from pydantic import Field, model_validator

from gen_epix.filter.base import Filter
from gen_epix.filter.enum import FilterType


class UuidSetFilter(Filter):
    members: frozenset[UUID] = Field(
        default=None, description="The UUIDs to match.", frozen=True
    )

    @model_validator(mode="after")
    def _validate_state(self) -> Self:
        self._match = lambda x: x in self.members  # type: ignore
        return self

    def _match(self, value: Any) -> bool:
        # Function is implemented dynamically in _validate_state
        raise NotImplementedError()


class TypedUuidSetFilter(UuidSetFilter):
    type: Literal[FilterType.UUID_SET.value]



from typing import Any
from uuid import UUID

from gen_epix.commondb.app_impl_details import AppImplDetails
from gen_epix.commondb.domain import command, exc, model
from gen_epix.commondb.domain.policy import BaseReadUserPolicy
from gen_epix.commondb.domain.service import BaseAbacService
from gen_epix.fastapp import Command
from gen_epix.fastapp.enum import CrudOperation, CrudOperationSet
from gen_epix.filter.composite import CompositeFilter
from gen_epix.filter.enum import LogicalOperator
from gen_epix.filter.equals_boolean import EqualsBooleanFilter
from gen_epix.filter.equals_uuid import EqualsUuidFilter


class ReadUserPolicy(BaseReadUserPolicy):
    # TODO: replace by get_content implementation for more efficient application DURING execution

    def __init__(self, abac_service: BaseAbacService, **kwargs: Any):
        super().__init__(abac_service, **kwargs)

        app_impl: AppImplDetails = abac_service.app.impl
        self.organization_admin_policy_crud_command_class: type[
            command.OrganizationAdminPolicyCrudCommand
        ] = app_impl.get_mapped_class(command.OrganizationAdminPolicyCrudCommand)
        self.role_map = app_impl.role_map
        self.role_set_map = app_impl.role_set_map

    def filter(
        self, cmd: Command, results: model.User | list[model.User]
    ) -> model.User | list[model.User]:
        if not isinstance(cmd, command.UserCrudCommand):
            raise NotImplementedError(
                "Unsupported command type: {cmd.__class__.__name__}"
            )
        if cmd.operation not in CrudOperationSet.READ.value:
            # Not applicable
            return results
        user: model.User | None = cmd.user
        if user is None or user.id is None:
            raise AssertionError("User must be authenticated")
        is_no_abac_user = (
            len(
                user.roles.intersection(
                    self.role_set_map[model.enum.RoleSet.GE_APP_ADMIN]
                )
            )
            > 0
        )
        if is_no_abac_user:
            return results

        organization_ids: set[UUID]
        org_admin_policies: list[model.OrganizationAdminPolicy]
        is_org_admin = (
            len(
                user.roles.intersection(
                    self.role_set_map[model.enum.RoleSet.GE_ORG_ADMIN]
                )
            )
            > 0
        )
        if is_org_admin:
            # User is organization admin: can read all users (active or not) of all
            # organizations they are admin of, plus all organization admins of those
            org_admin_policies = self.abac_service.app.handle(
                self.organization_admin_policy_crud_command_class(
                    user=user,
                    operation=CrudOperation.READ_ALL,
                    query_filter=EqualsBooleanFilter(key="is_active", value=True),
                )
            )
            organization_ids = {
                x.organization_id for x in org_admin_policies if x.user_id == user.id
            }
            org_admin_user_ids = {
                x.user_id
                for x in org_admin_policies
                if x.organization_id in organization_ids
            }
        else:
            # Regular user: can read only self and active organization admins of own organization
            org_admin_policies = self.abac_service.app.handle(
                self.organization_admin_policy_crud_command_class(
                    user=user,
                    operation=CrudOperation.READ_ALL,
                    query_filter=CompositeFilter(
                        filters=[
                            EqualsUuidFilter(
                                key="organization_id", value=user.organization_id
                            ),
                            EqualsBooleanFilter(key="is_active", value=True),
                        ],
                        operator=LogicalOperator.AND,
                    ),
                )
            )
            organization_ids = set()
            org_admin_user_ids = {x.user_id for x in org_admin_policies}
        user_ids = {user.id} | org_admin_user_ids

        # Filter or check results
        result_list: list[model.User]
        if cmd.operation == CrudOperation.READ_ALL:
            # Open-ended results: filter
            result_list = results  # type:ignore[assignment]
            return [
                x
                for x in result_list
                if (x.id in user_ids or x.organization_id in organization_ids)
                and (x.is_active or is_org_admin)
            ]
        elif cmd.operation == CrudOperation.READ_SOME:
            # Specific users requested: check results
            result_list = results  # type:ignore[assignment]
            if not all(
                (x.organization_id in organization_ids or x.id in user_ids)
                and (x.is_active or is_org_admin)
                for x in result_list
            ):
                # User cannot read users outside their admin organizations
                raise exc.UnauthorizedAuthError(
                    "Cannot read users outside your admin organizations"
                )
            return results
        elif cmd.operation == CrudOperation.READ_ONE:
            # Specific user requested: check results
            result: model.User = results  # type:ignore[assignment]
            if (
                result.organization_id not in organization_ids
                and result.id not in user_ids
            ) or not (result.is_active or is_org_admin):
                # User cannot read users outside their admin organizations
                raise exc.UnauthorizedAuthError(
                    "Cannot read users outside your admin organizations"
                )
            return results
        raise NotImplementedError("Unsupported operation: {cmd.operation.value}")



import json
from collections.abc import Iterable
from datetime import datetime
from logging import Logger
from typing import Any

import httpx
import jwt

from gen_epix.fastapp import HttpProtocol, RemoteApp, exc
from gen_epix.fastapp.enum import AuthProtocol, OAuthFlow
from gen_epix.fastapp.log import LogItem
from gen_epix.fastapp.model import Command
from gen_epix.fastapp.services.auth.model import OidcServerCfg
from gen_epix.fastapp.services.auth.oauth_idp_client import OauthIdpClient
from gen_epix.seqdb.api import RetrievePhylogeneticTreeRequestBody
from gen_epix.seqdb.domain import DOMAIN
from gen_epix.seqdb.domain import command as seqdb_command
from gen_epix.seqdb.domain import model as seqdb_model


class SeqdbRemoteApp(RemoteApp):

    DEFAULT_ROUTE_PREFIX = "/v1/"

    DEFAULT_OAUTH_TOKEN_REFRESH_MARGIN = 60  # seconds

    ROUTE_MAP: dict[type[Command], str] = {
        seqdb_command.RetrievePhylogeneticTreeCommand: "retrieve/phylogenetic_tree",
        seqdb_command.RetrieveSeqFastaCommand: "retrieve/genetic_sequence/fasta",
    }

    def __init__(
        self,
        host: str,
        port: int | None,
        http_protocol: HttpProtocol = HttpProtocol.HTTPS,
        default_route_prefix: str | None = None,
        default_headers: dict[str, str] | None = None,
        ssl_cert_file: str | None = None,
        auth_protocol: AuthProtocol | str = AuthProtocol.NONE,
        oauth_flow: OAuthFlow | str | None = None,
        oauth_discovery_url: str | None = None,
        oauth_client_id: str | None = None,
        oauth_client_secret: str | None = None,
        oauth_scope: str | None = None,
        oauth_token_endpoint: str | None = None,
        oauth_token_refresh_margin: float | None = None,
        logger: Logger | None = None,
        log_item_class: type[LogItem] = LogItem,
        **kwargs: Any,
    ) -> None:
        if isinstance(auth_protocol, str):
            auth_protocol = AuthProtocol(auth_protocol)
        if isinstance(oauth_flow, str):
            oauth_flow = OAuthFlow(oauth_flow)
        default_route_prefix = default_route_prefix or self.DEFAULT_ROUTE_PREFIX
        oauth_token_refresh_margin = (
            oauth_token_refresh_margin or self.DEFAULT_OAUTH_TOKEN_REFRESH_MARGIN
        )

        super().__init__(
            DOMAIN,
            host,
            port,
            http_protocol=http_protocol,
            default_route_prefix=default_route_prefix,
            default_headers=default_headers,
            add_generated_crud_route_handlers=True,
            ssl_cert_file=ssl_cert_file,
            **kwargs,
        )

        # Initialize IDP client if needed
        oauth_idp_client: OauthIdpClient | None = None
        if auth_protocol == AuthProtocol.NONE:
            pass
        elif auth_protocol == AuthProtocol.OAUTH2:
            if oauth_discovery_url is None:
                raise exc.InitializationServiceError(
                    "OAuth discovery endpoint must be provided for OAUTH2 auth protocol"
                )
            if oauth_client_id is None:
                raise exc.InitializationServiceError(
                    "OAuth client ID must be provided for OAUTH2 auth protocol"
                )
            if oauth_scope is None:
                raise exc.InitializationServiceError(
                    "OAuth scope must be provided for OAUTH2 auth protocol"
                )
            oauth_idp_client = OauthIdpClient(
                server_cfg=OidcServerCfg(
                    name="",
                    label="",
                    discovery_url=oauth_discovery_url,
                    client_id=oauth_client_id,
                    client_secret=oauth_client_secret,
                    token_endpoint=oauth_token_endpoint,
                    scope=oauth_scope,
                ),
                ssl_context=self.ssl_context,
                logger=logger,
                log_item_class=log_item_class,
            )
        else:
            raise exc.InitializationServiceError(
                f"Auth protocol {auth_protocol} not supported"
            )
        self._auth_protocol = auth_protocol
        self._oauth_flow = oauth_flow
        self._oauth_idp_client = oauth_idp_client
        self._oauth_scope = oauth_scope
        self._oauth_token_refresh_margin = oauth_token_refresh_margin
        self._oauth_header_cache: tuple[int, dict[str, str]] | None = None

        # Register routes and handlers
        self.register_route(
            seqdb_command.RetrievePhylogeneticTreeCommand,
            self.ROUTE_MAP[seqdb_command.RetrievePhylogeneticTreeCommand],
        )
        self.register_route(
            seqdb_command.RetrieveSeqFastaCommand,
            self.ROUTE_MAP[seqdb_command.RetrieveSeqFastaCommand],
        )
        self.register_handler(
            seqdb_command.RetrievePhylogeneticTreeCommand,
            self.retrieve_phylogenetic_tree,
        )
        self.register_handler(
            seqdb_command.RetrieveSeqFastaCommand,
            self.retrieve_genetic_sequence_fasta_by_id,
        )

    def get_headers(self, cmd: Command) -> dict[str, str]:
        # Call identity provider to get JWT
        if self._auth_protocol == AuthProtocol.NONE:
            return self._default_headers
        if self._auth_protocol == AuthProtocol.OAUTH2:
            assert self._oauth_idp_client is not None
            assert self._oauth_scope is not None
            # Check if cached token is still valid
            if self._oauth_header_cache and self._oauth_header_cache[0] > (
                datetime.now().timestamp() - self._oauth_token_refresh_margin
            ):
                # Return cached header
                return self._oauth_header_cache[1]
            # Retrieve new token

            jwt_token = self._oauth_idp_client.retrieve_jwt_with_client_credentials_flow(
                scope=self._oauth_scope
            )
            # Create headers
            headers = dict(self._default_headers)
            headers["Authorization"] = f"Bearer {jwt_token}"
            # Put header in cache together with its expiry time
            claims = jwt.decode(jwt_token, options={"verify_signature": False})
            exp: int | None = claims.get("exp")
            if exp is None:
                # No expiration claim, valid forever
                self._oauth_header_cache = (int(datetime.max.timestamp()), headers)
            else:
                self._oauth_header_cache = (exp, headers)
            return headers
        raise exc.InitializationServiceError(
            f"Auth protocol {self._auth_protocol.value} not supported for token retrieval"
        )

    def retrieve_phylogenetic_tree(
        self,
        cmd: seqdb_command.RetrievePhylogeneticTreeCommand,
    ) -> seqdb_model.PhylogeneticTree | None:
        headers = self.get_headers(cmd)
        route = self.get_route(cmd)

        # Create request body matching seqdb API expectations

        request_body = RetrievePhylogeneticTreeRequestBody(
            seq_distance_protocol_id=cmd.seq_distance_protocol_id,
            tree_algorithm=cmd.tree_algorithm,
            seq_ids=cmd.seq_ids,
            leaf_codes=cmd.leaf_names,
        )

        with httpx.Client(verify=self.ssl_context) as client:
            response = client.post(
                route,
                json=json.loads(request_body.model_dump_json()),
                headers=headers,
            )
            response.raise_for_status()
            data = response.json()
        if not data:
            return None
        return seqdb_model.PhylogeneticTree(**data)

    def retrieve_genetic_sequence_fasta_by_id(
        self,
        cmd: seqdb_command.RetrieveSeqFastaCommand,
    ) -> Iterable[str]:
        headers = self.get_headers(cmd)
        route = self.get_route(cmd)

        request_body: dict[str, Any] = {
            "user": cmd.user,
            "seq_ids": cmd.seq_ids,
            "wrap": cmd.wrap,
        }

        def _iter_fasta_generator() -> Iterable[str]:
            with httpx.Client(verify=self.ssl_context) as client:
                with client.stream(
                    "POST", route, json=request_body, headers=headers
                ) as resp:
                    resp.raise_for_status()
                    for chunk in resp.iter_bytes():
                        yield chunk.decode()

        return _iter_fasta_generator()



"""
Pipeline for chaining transformers with error tracking and recovery mechanisms.
"""

import logging
import time
from collections.abc import Callable, Iterator
from typing import Any

from gen_epix.transform.stream_processer import StreamProcessor
from gen_epix.transform.transform_result import TransformResult
from gen_epix.transform.transformer import Transformer


class Pipeline(StreamProcessor):
    """Chainable pipeline of transformers with comprehensive error handling."""

    def __init__(self, transformers: list[Transformer] | None = None):
        self.transformers = transformers or []
        self.error_handlers: dict[str, Callable[[TransformResult], None]] = {}
        self.logger = logging.getLogger(__name__)

    def add(self, transformer: Transformer) -> "Pipeline":
        """Add transformer to pipeline."""
        self.transformers.append(transformer)
        return self

    def __or__(self, other: Transformer) -> "Pipeline":
        """Enable chaining with | operator."""
        return self.add(other)

    def register_error_handler(
        self, transformer_name: str, handler: Callable[[TransformResult], None]
    ) -> "Pipeline":
        """Register error handler for specific transformer."""
        self.error_handlers[transformer_name] = handler
        return self

    def process_stream(self, stream: Iterator[Any]) -> Iterator[TransformResult]:
        """Process stream through pipeline with error tracking."""
        for obj in stream:
            yield from self._process_single_object(obj)

    def _process_single_object(self, obj: Any) -> Iterator[TransformResult]:
        """Process single object through entire pipeline."""
        current_obj = obj

        for i, transformer in enumerate(self.transformers):
            try:
                result = transformer(current_obj)

                if not result.success:
                    # Handle transformation error
                    self._handle_error(result)
                    yield result
                    return  # Stop pipeline on error

                current_obj = result.transformed_object

                # Yield intermediate results if needed
                if i == len(self.transformers) - 1:  # Last transformer
                    yield result

            except Exception as e:
                error_result = TransformResult(
                    success=False,
                    original_object=obj,
                    error=e,
                    transformer_name=transformer.name,
                    stage=f"pipeline_stage_{i}",
                )
                self._handle_error(error_result)
                yield error_result
                return

    def _handle_error(self, result: TransformResult) -> None:
        """Handle transformation errors."""
        self.logger.error(
            f"Transformation failed in {result.transformer_name}: {result.error}"
        )

        # Call registered error handler if available
        if result.transformer_name and result.transformer_name in self.error_handlers:
            self.error_handlers[result.transformer_name](result)


class RetryTransformer(Transformer):
    """Wrapper that adds retry logic to any transformer."""

    def __init__(
        self,
        transformer: Transformer,
        max_retries: int = 3,
        backoff_factor: float = 1.0,
        name: str | None = None,
    ):
        super().__init__(name or f"Retry_{transformer.name}")
        self.transformer = transformer
        self.max_retries = max_retries
        self.backoff_factor = backoff_factor

    def transform(self, obj: Any) -> Any:
        """Transform with retry logic."""
        last_exception = None

        for attempt in range(self.max_retries + 1):
            try:
                return self.transformer.transform(obj)
            except Exception as e:
                last_exception = e
                if attempt < self.max_retries:
                    time.sleep(self.backoff_factor * (2**attempt))
                    continue
                break

        raise last_exception  # type: ignore[misc]


class FallbackTransformer(Transformer):
    """Use fallback transformer if primary fails."""

    def __init__(
        self, primary: Transformer, fallback: Transformer, name: str | None = None
    ):
        super().__init__(name)
        self.primary = primary
        self.fallback = fallback

    def transform(self, obj: Any) -> Any:
        """Transform with fallback on failure."""
        try:
            return self.primary.transform(obj)
        except Exception:
            return self.fallback.transform(obj)



# pylint: disable=useless-import-alias
from gen_epix import fastapp
from gen_epix.commondb.domain import enum as common_enum
from gen_epix.commondb.domain import model as common_model
from gen_epix.commondb.domain.model import (
    SORTED_MODELS_BY_SERVICE_TYPE as _COMMON_SORTED_MODELS_BY_SERVICE_TYPE,
)
from gen_epix.commondb.domain.model import Contact as Contact
from gen_epix.commondb.domain.model import DataCollection as DataCollection
from gen_epix.commondb.domain.model import DataCollectionSet as DataCollectionSet
from gen_epix.commondb.domain.model import (
    DataCollectionSetMember as DataCollectionSetMember,
)
from gen_epix.commondb.domain.model import IdentifierIssuer as IdentifierIssuer
from gen_epix.commondb.domain.model import Model as Model
from gen_epix.commondb.domain.model import Organization as Organization
from gen_epix.commondb.domain.model import (
    OrganizationAdminPolicy as OrganizationAdminPolicy,
)
from gen_epix.commondb.domain.model import OrganizationSet as OrganizationSet
from gen_epix.commondb.domain.model import (
    OrganizationSetMember as OrganizationSetMember,
)
from gen_epix.commondb.domain.model import Outage as Outage
from gen_epix.commondb.domain.model import Site as Site
from gen_epix.commondb.domain.model import User as User
from gen_epix.commondb.domain.model import UserInvitation as UserInvitation
from gen_epix.commondb.domain.model import (
    UserInvitationConstraints as UserInvitationConstraints,
)
from gen_epix.commondb.domain.model import UserNameEmail as UserNameEmail
from gen_epix.fastapp.services.auth import IdentityProvider as IdentityProvider
from gen_epix.fastapp.services.auth import IDPUser as IDPUser
from gen_epix.seqdb.domain import enum
from gen_epix.seqdb.domain.model.file import File as File
from gen_epix.seqdb.domain.model.seq import AlignmentMixin as AlignmentMixin
from gen_epix.seqdb.domain.model.seq import AlignmentProtocol as AlignmentProtocol
from gen_epix.seqdb.domain.model.seq import Allele as Allele
from gen_epix.seqdb.domain.model.seq import AlleleAlignment as AlleleAlignment
from gen_epix.seqdb.domain.model.seq import AlleleProfile as AlleleProfile
from gen_epix.seqdb.domain.model.seq import AssemblyProtocol as AssemblyProtocol
from gen_epix.seqdb.domain.model.seq import AstMeasurement as AstMeasurement
from gen_epix.seqdb.domain.model.seq import AstPrediction as AstPrediction
from gen_epix.seqdb.domain.model.seq import AstProtocol as AstProtocol
from gen_epix.seqdb.domain.model.seq import CodeMixin as CodeMixin
from gen_epix.seqdb.domain.model.seq import (
    CompleteAlleleProfile as CompleteAlleleProfile,
)
from gen_epix.seqdb.domain.model.seq import CompleteContig as CompleteContig
from gen_epix.seqdb.domain.model.seq import CompleteSample as CompleteSample
from gen_epix.seqdb.domain.model.seq import CompleteSeq as CompleteSeq
from gen_epix.seqdb.domain.model.seq import CompleteSnpProfile as CompleteSnpProfile
from gen_epix.seqdb.domain.model.seq import ContigAlignment as ContigAlignment
from gen_epix.seqdb.domain.model.seq import (
    KmerDetectionProtocol as KmerDetectionProtocol,
)
from gen_epix.seqdb.domain.model.seq import KmerProfile as KmerProfile
from gen_epix.seqdb.domain.model.seq import LibraryPrepProtocol as LibraryPrepProtocol
from gen_epix.seqdb.domain.model.seq import Locus as Locus
from gen_epix.seqdb.domain.model.seq import (
    LocusDetectionProtocol as LocusDetectionProtocol,
)
from gen_epix.seqdb.domain.model.seq import LocusSet as LocusSet
from gen_epix.seqdb.domain.model.seq import LocusSetMember as LocusSetMember
from gen_epix.seqdb.domain.model.seq import MultipleAlignment as MultipleAlignment
from gen_epix.seqdb.domain.model.seq import PcrMeasurement as PcrMeasurement
from gen_epix.seqdb.domain.model.seq import PcrProtocol as PcrProtocol
from gen_epix.seqdb.domain.model.seq import PhylogeneticTree as PhylogeneticTree
from gen_epix.seqdb.domain.model.seq import ProtocolMixin as ProtocolMixin
from gen_epix.seqdb.domain.model.seq import QualityMixin as QualityMixin
from gen_epix.seqdb.domain.model.seq import RawSeq as RawSeq
from gen_epix.seqdb.domain.model.seq import ReadSet as ReadSet
from gen_epix.seqdb.domain.model.seq import RefAllele as RefAllele
from gen_epix.seqdb.domain.model.seq import RefSeq as RefSeq
from gen_epix.seqdb.domain.model.seq import RefSnp as RefSnp
from gen_epix.seqdb.domain.model.seq import RefSnpSet as RefSnpSet
from gen_epix.seqdb.domain.model.seq import RefSnpSetMember as RefSnpSetMember
from gen_epix.seqdb.domain.model.seq import Sample as Sample
from gen_epix.seqdb.domain.model.seq import Seq as Seq
from gen_epix.seqdb.domain.model.seq import SeqAlignment as SeqAlignment
from gen_epix.seqdb.domain.model.seq import SeqCategory as SeqCategory
from gen_epix.seqdb.domain.model.seq import SeqCategorySet as SeqCategorySet
from gen_epix.seqdb.domain.model.seq import SeqClassification as SeqClassification
from gen_epix.seqdb.domain.model.seq import (
    SeqClassificationProtocol as SeqClassificationProtocol,
)
from gen_epix.seqdb.domain.model.seq import SeqDistance as SeqDistance
from gen_epix.seqdb.domain.model.seq import SeqDistanceProtocol as SeqDistanceProtocol
from gen_epix.seqdb.domain.model.seq import SeqMixin as SeqMixin
from gen_epix.seqdb.domain.model.seq import SeqTaxonomy as SeqTaxonomy
from gen_epix.seqdb.domain.model.seq import SnpDetectionProtocol as SnpDetectionProtocol
from gen_epix.seqdb.domain.model.seq import SnpProfile as SnpProfile
from gen_epix.seqdb.domain.model.seq import SubtypingScheme as SubtypingScheme
from gen_epix.seqdb.domain.model.seq import Taxon as Taxon
from gen_epix.seqdb.domain.model.seq import TaxonLocusLink as TaxonLocusLink
from gen_epix.seqdb.domain.model.seq import TaxonomyProtocol as TaxonomyProtocol
from gen_epix.seqdb.domain.model.seq import TaxonSet as TaxonSet
from gen_epix.seqdb.domain.model.seq import TaxonSetMember as TaxonSetMember
from gen_epix.seqdb.domain.model.seq import TreeAlgorithm as TreeAlgorithm
from gen_epix.seqdb.domain.model.seq import TreeAlgorithmClass as TreeAlgorithmClass

# List up model classes per service and sorted according to links topology
SORTED_MODELS_BY_SERVICE_TYPE: dict[enum.ServiceType, list[type[fastapp.Model]]] = (
    {  # pyright: ignore[reportAssignmentType]
        # Common models
        enum.ServiceType.AUTH: list(
            _COMMON_SORTED_MODELS_BY_SERVICE_TYPE[common_enum.ServiceType.AUTH]
        ),
        enum.ServiceType.SYSTEM: list(
            _COMMON_SORTED_MODELS_BY_SERVICE_TYPE[common_enum.ServiceType.SYSTEM]
        ),
        enum.ServiceType.RBAC: list(
            _COMMON_SORTED_MODELS_BY_SERVICE_TYPE[common_enum.ServiceType.RBAC]
        ),
        enum.ServiceType.ORGANIZATION: list(
            _COMMON_SORTED_MODELS_BY_SERVICE_TYPE[common_enum.ServiceType.ORGANIZATION]
        ),
        # Specific models
        enum.ServiceType.ABAC: list(
            _COMMON_SORTED_MODELS_BY_SERVICE_TYPE[common_enum.ServiceType.ABAC]
        )
        + [],
        enum.ServiceType.FILE: [File],
        enum.ServiceType.SEQ: [
            SubtypingScheme,
            Taxon,
            TaxonSet,
            TaxonSetMember,
            Locus,
            TaxonLocusLink,
            LocusSet,
            LocusSetMember,
            RefSeq,
            RefAllele,
            RefSnp,
            RefSnpSet,
            RefSnpSetMember,
            AlignmentProtocol,
            AssemblyProtocol,
            AstProtocol,
            KmerDetectionProtocol,
            LibraryPrepProtocol,
            LocusDetectionProtocol,
            PcrProtocol,
            SeqClassificationProtocol,
            SeqDistanceProtocol,
            SnpDetectionProtocol,
            TaxonomyProtocol,
            TreeAlgorithmClass,
            TreeAlgorithm,
            SeqCategorySet,
            SeqCategory,
            Sample,
            RawSeq,
            ReadSet,
            Seq,
            Allele,
            AlleleProfile,
            KmerProfile,
            SnpProfile,
            AstMeasurement,
            AstPrediction,
            PcrMeasurement,
            SeqAlignment,
            AlleleAlignment,
            ContigAlignment,
            SeqClassification,
            SeqDistance,
            SeqTaxonomy,
            MultipleAlignment,
            PhylogeneticTree,
            CompleteContig,
            CompleteSample,
            CompleteSeq,
            CompleteSnpProfile,
        ],
        enum.ServiceType.FILE: [File],
    }
)

SORTED_SERVICE_TYPES = tuple(SORTED_MODELS_BY_SERVICE_TYPE.keys())

COMMON_MODEL_MAP: dict[type[fastapp.Model], type[fastapp.Model]] = {
    common_model.User: User,
    common_model.UserInvitation: UserInvitation,
    common_model.UserInvitationConstraints: UserInvitationConstraints,
    common_model.OrganizationAdminPolicy: OrganizationAdminPolicy,
}



from gen_epix.casedb.domain.enum import ServiceType
from gen_epix.casedb.domain.repository import BaseSubjectRepository
from gen_epix.fastapp import BaseService


class BaseSubjectService(BaseService):
    SERVICE_TYPE = ServiceType.SUBJECT

    # Property overridden to provide narrower return value to support linter
    @property  # type: ignore
    def repository(self) -> BaseSubjectRepository:  # type: ignore
        return super().repository  # type: ignore

    def register_handlers(self) -> None:
        self.register_default_crud_handlers()



# pylint: disable=useless-import-alias
from gen_epix.casedb.services.seqdb.remote_app import SeqdbRemoteApp as SeqdbRemoteApp
from gen_epix.casedb.services.seqdb.service import SeqdbService as SeqdbService



from collections.abc import Iterable
from uuid import UUID

import numpy as np
import sqlalchemy as sa

from gen_epix.fastapp import BaseUnitOfWork, CrudOperation
from gen_epix.fastapp.repositories import SARepository, SAUnitOfWork
from gen_epix.seqdb.domain import model
from gen_epix.seqdb.domain.repository import BaseSeqRepository
from gen_epix.seqdb.repositories import sa_model


class SeqSARepository(SARepository, BaseSeqRepository):

    def get_distance_matrix_by_seq_ids(
        self,
        uow: BaseUnitOfWork,
        seq_distance_protocol_id: UUID,
        seq_ids: list[UUID],
    ) -> np.ndarray:
        raise NotImplementedError("Code to be converted to seqdb architecture")
        self.raise_on_duplicate_ids(seq_ids)
        seqs = self.crud(
            uow,
            None,
            model.SeqDistance,
            None,
            seq_ids,
            CrudOperation.READ_SOME,
        )
        id_to_idx_map = {x.id: i for i, x in enumerate(seqs)}
        n = len(seqs)
        distance_matrix = np.empty((n, n))
        distance_matrix[:] = np.nan
        for i in range(n):
            for id_, distance in seqs[i].distances.items():
                if id_ not in id_to_idx_map:
                    continue
                distance_matrix[id_to_idx_map[id_], i] = distance
            distance_matrix[i, i] = 0
        return distance_matrix

    def retrieve_seq_fasta(
        self,
        uow: BaseUnitOfWork,
        seq_ids: list[UUID],
    ) -> Iterable[tuple[str, str]]:
        self.raise_on_duplicate_ids(seq_ids)
        assert isinstance(uow, SAUnitOfWork)
        stmt = (
            sa.select(sa_model.Seq.id, sa_model.RawSeq.seq)
            .join(sa_model.RawSeq, sa_model.Seq.raw_seq_id == sa_model.RawSeq.id)
            .where(sa_model.Seq.id.in_(seq_ids))
        )

        result = uow.session.execute(stmt)
        for seq_id, raw_seq in result:
            print(f"Yielding seq_id={seq_id} with raw_seq length={len(raw_seq)}")
            yield (seq_id, raw_seq)



# pylint: disable=useless-import-alias
from gen_epix.fastapp.repositories.dict.repository import (
    DictRepository as DictRepository,
)
from gen_epix.fastapp.repositories.dict.unit_of_work import (
    DictUnitOfWork as DictUnitOfWork,
)



import abc
import ssl
import uuid
from uuid import UUID

from fastapi import Request
from fastapi.openapi.models import SecurityBase

from gen_epix.fastapp.services.auth.model import Claims, IdentityProvider


class IdpClient(abc.ABC):

    DEFAULT_TOKEN = "id_token"

    def __init__(
        self,
        scheme_name: str,
        token_name: str | None = None,
        id: UUID | None = None,
        ssl_context: ssl.SSLContext | bool = True,
        **kwargs: dict,
    ) -> None:
        self.id: UUID = id or uuid.uuid4()

        self.ssl_context = ssl_context

        # Set SecurityBase properties
        self.scheme_name = scheme_name
        self.token_name = token_name or self.DEFAULT_TOKEN
        self.model: SecurityBase  # type: ignore

    @abc.abstractmethod
    def get_identity_provider(self) -> IdentityProvider:
        raise NotImplementedError()

    @abc.abstractmethod
    async def get_claims_from_jwt(
        self, jwt_token: str
    ) -> dict[str, str | int | bool | list[str]] | None:
        raise NotImplementedError()

    # TODO: make async
    @abc.abstractmethod
    def get_claims_from_userinfo(
        self, access_token: str
    ) -> dict[str, str | int | bool | list[str]]:
        raise NotImplementedError()

    @abc.abstractmethod
    async def __call__(self, request: Request) -> Claims | None:
        """
        Returns the claims of the user from the request or None if claims cannot be
        processed by this client.
        """
        raise NotImplementedError()
